# -*- coding: utf-8 -*-
from setuptools import setup

modules = \
['sutoppu']
setup_kwargs = {
    'name': 'sutoppu',
    'version': '0.1.2',
    'description': 'A simple python implementation of Specification pattern.',
    'long_description': "# Sutoppu\n\n[![Pypi Version](https://img.shields.io/pypi/v/sutoppu.svg)](https://pypi.org/project/sutoppu/)\n[![Python Version](https://img.shields.io/pypi/pyversions/sutoppu)](https://pypi.org/project/sutoppu/)\n[![CI](https://github.com/u8slvn/sutoppu/actions/workflows/ci.yml/badge.svg)](https://github.com/u8slvn/sutoppu/actions/workflows/ci.yml)\n[![Coverage Status](https://coveralls.io/repos/github/u8slvn/sutoppu/badge.svg?branch=master)](https://coveralls.io/github/u8slvn/sutoppu?branch=master)\n[![Project license](https://img.shields.io/pypi/l/sutoppu)](https://pypi.org/project/sutoppu/)\n[![Code Style](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\n**Sutoppu** (ストップ - Japanese from English *Stop*) is a simple python implementation of Specification pattern.\n\n## What is Specification Pattern?\n\nSee [Wikipedia](https://en.wikipedia.org/wiki/Specification_pattern).\n\n> In computer programming, the specification pattern is a particular software design pattern, whereby business rules can be recombined by chaining the business rules together using boolean logic. The pattern is frequently used in the context of domain-driven design.\n\nMore information: [Eric Evans and Martin Fowler article about Specifications](https://www.martinfowler.com/apsupp/spec.pdf)\n\n## Basic usage\n\n### Installation\n\n```sh\n$ pip install sutoppu\n```\n\n### Usage\n\n```python\nfrom sutoppu import Specification\n\n\nclass Fruit:\n    def __init__(self, color: str, sweet: bool, bitter: bool):\n        self.color = color\n        self.sweet = sweet\n        self.bitter = bitter\n\n\nclass FruitIsBitter(Specification):\n    description = 'The given fruit must be bitter.'\n\n    def is_satisfied_by(self, fruit: Fruit):\n        return fruit.bitter is True\n\n\nclass FruitIsSweet(Specification):\n    description = 'The given fruit must be sweet.'\n    \n    def is_satisfied_by(self, fruit: Fruit):\n        return fruit.sweet is True\n\n\nclass FruitIsYellow(Specification):\n    description = 'The given fruit must be yellow.'\n\n    def is_satisfied_by(self, fruit: Fruit):\n        return self.color == 'yellow'\n```\n\n```python\n>>> lemon = Fruit(color='yellow', sweet=False, bitter=True)\n>>> is_a_lemon = FruitIsYellow() & FruitIsBitter() & ~FruitIsSweet()\n>>> is_a_lemon.is_satisfied_by(lemon)\nTrue\n```\n\n### Operators\n\nBitwise operators are overload to provide simple syntax.\n\nAnd:\nyellow\n```python\n>>> my_spec = SpecificationA() & SpecificationB()\n```\n\nOr:\n\n```python\n>>> my_spec = SpecificationA() | SpecificationB()\n```\n\nNot:\n\n```python\n>>> my_spec = ~SpecificationA()\n```\n\n### Lighter syntax\n\nIf you do not find the `is_satisfied_by` method very convenient you can also directly call the specification as below.\n\n```python\n>>> lemon = Fruit(color='yellow', sweet=False, bitter=True)\n>>> is_a_lime = FruitIsGreen() & FruitIsBitter() & ~FruitIsSweet()\n>>> is_a_lime(lemon)\nFalse\n```\n\n### Error reporting\n\nIt can be difficult to know which specification failed in a complex rule. Sutoppu allows to list all the failed specifications by getting the `errors` attribute after use.\nThe `errors` attribute is reset each time the specification is used. For each failed specification, it returns a dict with the name of the specification class for key and the description provide in the class for value. In the case where the specification failed with a `not` condition, the description is prefixed with `Not ~`.\n\n```python\n>>> apple = Fruit(color='red', sweet=True, bitter=False)\n>>> is_a_lemon = FruitIsYellow() & FruitIsBitter() & ~ FruitIsSweet()\n>>> is_a_lemon.is_satisfied_by(apple)\nFalse\n>>> is_a_lemon.errors\n{\n    'FruitIsColored': 'The given fruit must be yellow.',\n    'FruitIsBitter': 'The given fruit must be bitter.',\n    'FruitIsSweet': 'Not ~ The given fruit must be sweet.'\n}\n```\n",
    'author': 'u8slvn',
    'author_email': 'u8slvn@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/u8slvn/sutoppu',
    'py_modules': modules,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)



"""Sutoppu
A simple implementation of Specification pattern.

Copyright (C) 2019 - u8slvn - Sylvain Collas
This module is released under the MIT License:
http://www.opensource.org/licenses/mit-license.php
"""

from abc import ABCMeta, abstractmethod
import functools

__all__ = ["Specification"]
__version__ = "0.1.2"


class _SpecificationMeta(ABCMeta):
    """Add a little bit of magic, _SpecificationMeta automatically apply the
    class method '_report_errors' as decorator for the 'is_satisfied_by'
    method. It allows to simplify Specification declaration by declaring only
    'is_satisfied_by' without paying attention of the '_report_errors'
    decorator.
    """

    def __new__(mcs, name, bases, namespace):
        cls = super().__new__(mcs, name, bases, namespace)
        if hasattr(cls, "is_satisfied_by") and hasattr(cls, "_report_errors"):
            cls.is_satisfied_by = cls._report_errors(cls.is_satisfied_by)
        return cls


class Specification(metaclass=_SpecificationMeta):
    """Specification base class, each domain specification must inherit from
    this class.
    """

    description = "No description provided."

    def __init__(self):
        self.errors = {}

    @classmethod
    def _report_errors(cls, func):
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            self.errors = {}  # reset the errors dict
            result = func(self, *args, **kwargs)
            self._report_error(result)
            return result

        return wrapper

    @abstractmethod
    def is_satisfied_by(self, candidate: any) -> bool:
        raise NotImplementedError

    def _report_error(self, result: bool):
        """Each time a specification verification failed, this method
        report it in the 'errors' dict attribute.
        """
        if not result:
            self.errors.update({self.class_name: self.description})

    @property
    def class_name(self):
        return self.__class__.__name__

    def __and__(self, spec: "Specification") -> "_AndSpecification":
        return _AndSpecification(self, spec)

    def __or__(self, spec: "Specification") -> "_OrSpecification":
        return _OrSpecification(self, spec)

    def __invert__(self) -> "_NotSpecification":  # not
        return _NotSpecification(self)

    def __call__(self, candidate: any):
        """Extra syntax for more facilities."""
        return self.is_satisfied_by(candidate)

    def __repr__(self):
        return f"<{self.class_name}: {self.description}>"


class _AndOrSpecification(Specification):
    """Base class for 'And' and 'Or' specifications."""

    def __init__(self, spec_a: Specification, spec_b: Specification):
        super().__init__()
        self._specs = (spec_a, spec_b)

    def _report_error(self, _):
        """Gets the children spec errors and merge them into its own.
        The goal behind this it to propagate the errors through all the
        parents specifications.
        The result argument is useless in this case.
        """
        for spec in self._specs:
            self.errors = {**self.errors, **spec.errors}

    def is_satisfied_by(self, candidate: any) -> bool:
        results = (spec.is_satisfied_by(candidate) for spec in self._specs)

        return self._check(*results)

    @abstractmethod
    def _check(self, spec_a: bool, spec_b: bool):
        """Check the operator logic."""
        raise NotImplementedError


class _AndSpecification(_AndOrSpecification):
    def _check(self, spec_a: bool, spec_b: bool) -> bool:
        return spec_a and spec_b


class _OrSpecification(_AndOrSpecification):
    def _check(self, spec_a: bool, spec_b: bool) -> bool:
        return spec_a or spec_b


class _NotSpecification(Specification):
    def __init__(self, spec: Specification):
        super().__init__()
        self._spec = spec

    def _report_error(self, result: bool):
        """Due to its inversion logic the not specification must report
        an error if the checked specification did not failed.
        The description is prefixed with 'Not ~'  to indicate this.
        """
        if not result:
            description = f"Not ~ {self._spec.description}"
            self.errors.update({self._spec.class_name: description})

    def is_satisfied_by(self, candidate: any) -> bool:
        result = self._spec.is_satisfied_by(candidate)

        return not result
